#!/usr/bin/env python

# Built-in
import sys
import os


# Generic
import matplotlib.pyplot as plt
plt.switch_backend('Qt5Agg')
plt.ioff()


# tofu
# test if in a tofu git repo
_HERE = os.path.abspath(os.path.dirname(__file__))
_TOFUPATH = os.path.dirname(os.path.dirname(_HERE))
istofugit = False
if '.git' in os.listdir(_TOFUPATH) and 'tofu' in _TOFUPATH:
    istofugit = True

if istofugit:
    # Make sure we load the corresponding tofu
    sys.path.insert(1, _TOFUPATH)
    import tofu as tf
    _ = sys.path.pop(1)
else:
    import tofu as tf


# import parser dict
sys.path.insert(1, _TOFUPATH)
from scripts._dparser import _DPARSER
_ = sys.path.pop(1)


# tforigin = tf.__file__
# tfversion = tf.__version__
# print(tforigin, tfversion)


if 'imas2tofu' not in dir(tf):
    msg = "imas does not seem to be available\n"
    msg += "  => tf.imas2tofu not available\n"
    msg += "  => tofuplot not available"
    raise Exception(msg)


###################################################
###################################################
#       default values
###################################################


_DCONVERT = {
    'background': 'bck'
}


###################################################
###################################################
#       function
###################################################


def call_tfcalcimas(shot=None, run=None, user=None,
                    tokamak=None, version=None,
                    tokamak_eq=None, user_eq=None,
                    shot_eq=None, run_eq=None,
                    tokamak_prof=None, user_prof=None,
                    shot_prof=None, run_prof=None,
                    ids=None, t0=None, tlim=None, extra=None,
                    plot_compare=True, Brightness=None,
                    res=None, interp_t=None, coefs=None,
                    sharex=None, indch=None, indch_auto=None,
                    input_file=None, output_file=None,
                    background=None, ddef=None):

    # --------------
    # Check inputs
    kwd = locals()
    for k0 in set(ddef.keys()).intersection(kwd.keys()):
        if kwd[k0] is None:
            kwd[k0] = ddef[k0]
    for k0 in set(_DCONVERT.keys()).intersection(kwd.keys()):
        kwd[_DCONVERT[k0]] = kwd[k0]
        del kwd[k0]
    del kwd['ddef']

    if isinstance(kwd['t0'], str) and kwd['t0'].lower() == 'none':
        kwd['t0'] = None
    if kwd['tlim'] is not None and len(kwd['tlim']) == 1:
        kwd['tlim'] = [kwd['tlim'], None]
    if kwd['tlim'] is not None and len(kwd['tlim']) != 2:
        msg = ("tlim must contain 2 limits:\n"
               + "\t- provided: {}".format(kwd['tlim']))
        raise Exception(msg)

    # --------------
    # run
    tf.calc_from_imas(plot=None, **kwd)
    plt.show(block=True)


###################################################
###################################################
#       bash call (main)
###################################################


# if __name__ == '__main__':
def main():

    # Instanciate parser
    ddef, parser = _DPARSER['calc']()

    # Parse arguments
    args = parser.parse_args()

    # Call wrapper function
    call_tfcalcimas(ddef=ddef, **dict(args._get_kwargs()))


# Add this to make sure it remains executable even without install
if __name__ == '__main__':
    main()




# Built-in
import os
import warnings
import functools as ftools

# Common
import numpy as np

# tofu
pfe = os.path.join(os.path.expanduser('~'), '.tofu', '_imas2tofu_def.py')
if os.path.isfile(pfe):
    # Make sure we load the user-specific file
    # sys.path method
    # sys.path.insert(1, os.path.join(os.path.expanduser('~'), '.tofu'))
    # import _scripts_def as _defscripts
    # _ = sys.path.pop(1)
    # importlib method
    import importlib.util
    spec = importlib.util.spec_from_file_location("_defimas2tofu", pfe)
    _defimas2tofu = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(_defimas2tofu)
else:
    try:
        import tofu.imas2tofu._def as _defimas2tofu
    except Exception as err:
        from . import _def as _defimas2tofu

# imas
try:
    import imas
except Exception as err:
    raise Exception('imas not available')


_DSHORT = _defimas2tofu._dshort
_DCOMP = _defimas2tofu._dcomp
_DDUNITS = imas.dd_units.DataDictionaryUnits()

_ISCLOSE = True
_POS = False
_EMPTY = True
_NAN = True
_DATA = True
_UNITS = True
_STRICT = True
_STACK = True

# #############################################################################
#                      Units functions
# #############################################################################


def _prepare_sig_units(sig, units=False):
    """ Remove brackest from shortcut, if any """
    if '[' in sig:
        while '[' in sig:
            sig = sig[:sig.index('[')] + sig[sig.index(']')+1:]
    return sig


def get_units(ids, sig, dshort=None, dcomp=None, force=None):
    """ Get units from imas.dd_units.DataDictionaryUnits() """
    if dshort is None:
        dshort = _DSHORT
    if dcomp is None:
        dcomp = _DCOMP
    if force is None:
        force = True
    if sig in dshort[ids].keys():
        sig = _prepare_sig_units(dshort[ids][sig]['str'])
    else:
        sig = _prepare_sig_units(sig)
    units = _DDUNITS.get_units(ids, sig.replace('.', '/'))

    # Condition in which to use tofu units instead of imas units
    c0 = (units is None
          and force is True
          and (sig in dshort[ids].keys() or sig in dcomp[ids].keys()))
    if c0 is True:
        if sig in dshort[ids].keys():
            tofuunits = dshort[ids][sig].get('units')
        else:
            tofuunits = dcomp[ids][sig].get('units')
        if tofuunits != units:
            units = tofuunits
    return units


# #############################################################################
#                      Data retrieveing function
# #############################################################################


def _prepare_sig(sig):
    if '[' in sig:
        # Get nb and ind
        ind0 = 0
        while '[' in sig[ind0:]:
            ind1 = ind0 + sig[ind0:].index('[')
            ind2 = ind0 + sig[ind0:].index(']')
            sig = sig.replace(sig[ind1+1:ind2],
                              sig[ind1+1:ind2].replace('.', '/'))
            ind0 = ind2 + 1
    return sig


def _get_condfromstr(sid, sig=None):
    lid0, id1 = sid.split('=')
    lid0 = lid0.split('.')

    if '.' in id1 and id1.replace('.', '').isdecimal():
        id1 = float(id1)
    elif id1.isdecimal():
        id1 = int(id1)
    elif '.' in id1:
        msg = ("Not clear how to interpret the following condition:\n"
               + "\t- sig: {}\n".format(sig)
               + "\t- condition: {}".format(sid))
        raise Exception(msg)
    return lid0, id1


def get_fsig(sig):
    # break sig in list of elementary nodes
    sig = _prepare_sig(sig)
    ls0 = sig.split('.')
    sig = sig.replace('/', '.')
    ls0 = [ss.replace('/', '.') for ss in ls0]
    ns = len(ls0)

    # For each node, identify type (i.e. [])
    lc = [all([si in ss for si in ['[', ']']]) for ss in ls0]
    dcond, seq, nseq, jj = {}, [], 0, 0
    for ii in range(0, ns):
        nseq = len(seq)
        if lc[ii]:
            # there is []
            if nseq > 0:
                dcond[jj] = {'type': 0, 'lstr': seq}
                seq = []
                jj += 1

            # Isolate [strin]
            ss = ls0[ii]
            strin = ss[ss.index('[')+1:-1]

            # typ 0 => no dependency
            # typ 1 => dependency ([],[time],[chan],[int])
            # typ 2 => selection ([...=...])
            cond, ind, typ = None, None, 1
            if '=' in strin:
                typ = 2
                cond = _get_condfromstr(strin, sig=sig)
            elif strin in ['time', 'chan']:
                ind = strin
            elif strin.isnumeric():
                ind = [int(strin)]
            dcond[jj] = {'str': ss[:ss.index('[')], 'type': typ,
                         'ind': ind, 'cond': cond}
            jj += 1
        else:
            seq.append(ls0[ii])
            if ii == ns-1:
                dcond[jj] = {'type': 0, 'lstr': seq}

    c0 = [v['type'] == 1 and (v['ind'] is None or len(v['ind']) > 1)
          for v in dcond.values()]
    if np.sum(c0) > 1:
        msg = ("Cannot handle mutiple iterative levels yet !\n"
               + "\t- sig: {}".format(sig))
        raise Exception(msg)

    # Create function for getting signal
    def fsig(obj, indt=None, indch=None, stack=None, dcond=dcond):
        if stack is None:
            stack = _STACK
        sig = [obj]
        nsig = 1
        for ii in dcond.keys():

            # Standard case (no [])
            if dcond[ii]['type'] == 0:
                sig = [ftools.reduce(getattr, [sig[jj]]+dcond[ii]['lstr'])
                       for jj in range(0, nsig)]

            # dependency
            elif dcond[ii]['type'] == 1:
                for jj in range(0, nsig):
                    sig[jj] = getattr(sig[jj], dcond[ii]['str'])
                    nb = len(sig[jj])
                    if dcond[ii]['ind'] == 'time':
                        ind = indt
                    elif dcond[ii]['ind'] == 'chan':
                        ind = indch
                    else:
                        ind = dcond[ii]['ind']

                    if ind is None:
                        ind = range(0, nb)
                    if nsig > 1:
                        assert type(ind) is not str and len(ind) == 1

                    if len(ind) == 1:
                        sig[jj] = sig[jj][ind[0]]
                    else:
                        assert nsig == 1
                        sig = [sig[0][ll] for ll in ind]
                        nsig = len(sig)

            # one index to be found
            else:
                for jj in range(0, nsig):
                    sig[jj] = getattr(sig[jj], dcond[ii]['str'])
                    nb = len(sig[jj])
                    typ = type(ftools.reduce(
                        getattr, [sig[jj][0]] + dcond[ii]['cond'][0]))
                    if typ == str:
                        ind = [
                            ll for ll in range(0, nb)
                            if (ftools.reduce(
                                getattr,
                                [sig[jj][ll]] + dcond[ii]['cond'][0]).strip()
                                == dcond[ii]['cond'][1].strip())]
                    else:
                        ind = [ll for ll in range(0, nb)
                               if (ftools.reduce(
                                   getattr,
                                   [sig[jj][ll]]+dcond[ii]['cond'][0])
                                   == dcond[ii]['cond'][1])]
                    if len(ind) != 1:
                        msg = ("No / several matching signals for:\n"
                               + "\t- {}[]{} = {}\n".format(
                                   dcond[ii]['str'],
                                   dcond[ii]['cond'][0],
                                   dcond[ii]['cond'][1])
                               + "\t- nb.of matches: {}".format(len(ind)))
                        raise Exception(msg)
                    sig[jj] = sig[jj][ind[0]]

        # Conditions for stacking / sqeezing sig
        lc = [(stack and nsig > 1 and isinstance(sig[0], np.ndarray)
               and all([ss.shape == sig[0].shape for ss in sig[1:]])),
              (stack and nsig > 1
               and type(sig[0]) in [int, float, np.int_, np.float_, str]),
              (stack and nsig == 1
               and type(sig) in [np.ndarray, list, tuple])]

        if lc[0]:
            sig = np.atleast_1d(np.squeeze(np.stack(sig)))
        elif lc[1] or lc[2]:
            sig = np.atleast_1d(np.squeeze(sig))
        return sig
    return fsig


# #############################################################################
#                      Data functions
# #############################################################################


def _checkformat_getdata_ids(ids=None, dids=None):
    """ Check the desired sids is available """
    if not isinstance(dids, dict):
        msg = ("dids must be a dict\n"
               + "\t- provided: {}".format(type(dids)))
        raise Exception(msg)

    msg = ("Arg ids must be either:\n"
           + "\t- None: if self.dids only has one key\n"
           + "\t- str: a valid key of self.dids\n\n"
           + "  Provided : {}\n".format(ids)
           + "  Available: {}\n".format(list(dids.keys()))
           + "  => Consider using self.add_ids({})".format(str(ids)))

    lc = [ids is None, type(ids) is str]
    if not any(lc):
        raise Exception(msg)

    if lc[0]:
        if len(dids.keys()) != 1:
            raise Exception(msg)
        ids = list(dids.keys())[0]
    elif lc[1]:
        if ids not in dids.keys():
            raise Exception(msg)
    return ids


def _checkformat_getdata_sig(sig, ids,
                             dshort=None, dcomp=None, dall_except=None):
    lc = [isinstance(dd, dict) for dd in [dshort, dcomp, dall_except]]
    if not all(lc):
        msg = ("dshort, dcomp and dall_except must be dict:\n"
               + "\t- type(dshort): {}\n".format(type(dshort))
               + "\t- type(dcomp): {}\n".format(type(dcomp))
               + "\t- type(dall_except): {}\n".format(type(dall_except)))
        raise Exception(msg)

    msg = ("Arg sig must be a str or a list of str!\n"
           + "  More specifically, a list of valid ids nodes paths")
    lks = list(dshort[ids].keys())
    lkc = list(dcomp[ids].keys())
    lk = set(lks).union(lkc)

    if ids in dall_except.keys():
        lk = lk.difference(dall_except[ids])

    lc = [sig is None, type(sig) is str, type(sig) is list]
    if not any(lc):
        raise Exception(msg)
    if lc[0]:
        sig = list(lk)
    elif lc[1]:
        sig = [sig]
    elif lc[2]:
        if any([type(ss) is not str for ss in sig]):
            raise Exception(msg)
    nsig = len(sig)

    # Check each sig is either a key / value[str] to dshort
    comp = np.zeros((nsig,), dtype=bool)
    for ii in range(0, nsig):
        lc0 = [sig[ii] in lks,
               [sig[ii] == dshort[ids][kk]['str'] for kk in lks]]
        c1 = sig[ii] in lkc
        if not (lc0[0] or any(lc0[1]) or c1):
            msg = ("Each provided sig must be either:\n"
                   + "\t- a valid shortcut (cf. self.shortcuts()\n"
                   + "\t- a valid long version (cf. self.shortcuts)\n"
                   + "\n  Provided sig: {} for ids {}".format(sig, ids))
            raise Exception(msg)
        if c1:
            comp[ii] = True
        else:
            if not lc0[0]:
                sig[ii] = lks[lc0[1].index(True)]
    return sig, comp


def _checkformat_getdata_occ(occ, ids, dids=None):
    if not isinstance(dids, dict):
        msg = ("dids must be a dict\n"
               + "\t- provided: {}".format(type(dids)))
        raise Exception(msg)
    msg = ("Arg occ must be a either:\n"
           + "\t- None: all occurences are used\n"
           + "\t- int: occurence (in self.dids[{}]['occ'])\n".format(ids)
           + "\t- int array: occurences (in self.dids[{}]['occ'])".format(ids))
    lc = [occ is None, type(occ) is int, hasattr(occ, '__iter__')]
    if not any(lc):
        raise Exception(msg)
    if lc[0]:
        occ = dids[ids]['occ']
    else:
        occ = np.r_[occ].astype(int).ravel()
        if any([oc not in dids[ids]['occ'] for oc in occ]):
            raise Exception(msg)
    return occ


def _checkformat_getdata_indch(indch, nch):
    msg = ("Arg indch must be a either:\n"
           + "    - None: all channels used\n"
           + "    - int: channel to use (index)\n"
           + "    - array of int: channels to use (indices)\n"
           + "    - array of bool: channels to use (indices)\n")
    lc = [indch is None,
          isinstance(indch, int),
          hasattr(indch, '__iter__') and not isinstance(indch, str)]
    if not any(lc):
        raise Exception(msg)
    if lc[0]:
        indch = np.arange(0, nch)
    elif lc[1] or lc[2]:
        indch = np.r_[indch].ravel()
        lc = [indch.dtype == np.int, indch.dtype == np.bool]
        if not any(lc):
            raise Exception(msg)
        if lc[1]:
            indch = np.nonzero(indch)[0]
        assert np.all((indch >= 0) & (indch < nch))
    return indch


def _check_data(data, pos=None, nan=None, isclose=None, empty=None):
    """ Check the data loaded from imas against several safety checks

    For each occurence

    Available checks:
        - only positive values
        - default IMAS values (>1e30)
        - duplicated unique vector
        - empty data (shape contains 0 oir size 0 or only nan)

    """

    # ------------
    # Check inputs
    if pos is None:
        pos = _POS
    if nan is None:
        nan = _NAN
    if isclose is None:
        isclose = _ISCLOSE
    if empty is None:
        empty = _EMPTY

    # ------------
    # Run checks on data

    # If isclose, check data contains a replicated vector (keep vector only)
    if isclose is True:
        for ii in range(0, len(data)):
            if isinstance(data[ii], np.ndarray) and data[ii].ndim == 2:
                if np.allclose(data[ii], data[ii][0:1, :]):
                    data[ii] = data[ii][0, :]
                elif np.allclose(data[ii], data[ii][:, 0:1]):
                    data[ii] = data[ii][:, 0]

    # All values larger than 1e30 are default imas values => nan
    if nan is True:
        for ii in range(0, len(data)):
            c0 = (isinstance(data[ii], np.ndarray)
                  and data[ii].dtype in [np.float_, np.int_])
            if c0 is True:
                # Make sure to test only non-nan to avoid warning
                ind = (~np.isnan(data[ii])).nonzero()
                ind2 = np.abs(data[ii][ind]) > 1.e30
                ind = tuple([ii[ind2] for ii in ind])
                data[ii][ind] = np.nan

    # data supposed to be positive only (nan otherwise)
    if pos is True:
        for ii in range(0, len(data)):
            if isinstance(data[ii], np.ndarray):
                data[ii][data[ii] < 0] = np.nan

    # data appears to be empty or all nan
    isempty = [None for ii in range(len(data))]
    if empty is True:
        for ii in range(len(data)):
            isempty[ii] = (len(data[ii]) == 0
                           or (isinstance(data[ii], np.ndarray)
                               and (data[ii].size == 0
                                    or 0 in data[ii].shape)))
            if isinstance(data[ii], np.ndarray) and data[ii].dtype.kind != 'U':
                isempty[ii] &= bool(np.all(np.isnan(data[ii])))
    return data, isempty


def _get_data_units(ids=None, sig=None, occ=None,
                    comp=None, indt=None, indch=None,
                    stack=None, isclose=None, flatocc=None,
                    nan=None, pos=None, empty=None, warn=None,
                    dids=None, dcomp=None, dshort=None, dall_except=None,
                    data=True, units=True):
    """ Reference method for getting data and units, using shortcuts

    For a given ids, sig (shortcut) and occurence (occ)
    Get data if data = True
    Get units if units = True

    The data is:
        - restricted to the desired time indices (indt)
        - restricted to the desired channels (indch)
        - stacked into a single array if stack = True
        - checked whether many vectors are duplicates (expected)
            in which case a single vector is kept
        - checked for NaNs if nan = True
        - checked for positive values if pos = True
        - All occurences are flatened if flatocc = True

    A warning is issued if warn = True

    If comp = True, it means the data is not loaded diurectly from the ids,
    but computed from a combination of signals in the ids.
    The computation function must have been defined previously

    """

    # --------------
    # Check inputs

    if data is None:
        data = _DATA
    if units is None:
        units = _UNITS

    # --------------
    # Load data

    # get list of results for occ
    occref = dids[ids]['occ']
    indoc = np.array([np.nonzero(occref == oc)[0][0] for oc in occ])
    nocc = len(indoc)
    out, unit, errdata, errunits = None, None, None, None

    # Data has te be computed
    if comp is True:
        if data is True:
            if stack is None:
                stack = True
            try:
                lstr = dcomp[ids][sig]['lstr']
                kargs = dcomp[ids][sig].get('kargs', {})
                ddata = get_data_units(
                    ids=ids, sig=lstr, occ=occ, indch=indch,
                    data=True, units=False, indt=indt, stack=stack,
                    flatocc=False, nan=nan, pos=pos, warn=warn,
                    dids=dids, dcomp=dcomp, dshort=dshort,
                    dall_except=dall_except)[0]
                out = [dcomp[ids][sig]['func'](
                    *[ddata[kk]['data'][nn] for kk in lstr],
                    **kargs)
                       for nn in range(0, nocc)]
                if pos is None:
                    pos = dcomp[ids][sig].get('pos', False)
            except Exception as err:
                errdata = str(err)

        if units is True:
            try:
                unit = dcomp[ids][sig].get('units', None)
            except Exception as err:
                errunits = str(err)

    # Data available from ids
    else:
        if data is True:
            try:
                out = [dshort[ids][sig]['fsig'](dids[ids]['ids'][ii],
                                                indt=indt, indch=indch,
                                                stack=stack)
                       for ii in indoc]
                if pos is None:
                    pos = dshort[ids][sig].get('pos', False)
            except Exception as err:
                errdata = str(err)

        if units is True:
            try:
                unit = get_units(ids, sig)
            except Exception as err:
                errunits = str(err)

    # Check data
    isempty = None
    if errdata is None and data is True:
        out, isempty = _check_data(out,
                                   pos=pos, nan=nan,
                                   isclose=isclose, empty=empty)
        if np.all(isempty):
            msg = ("empty data in {}.{}".format(ids, sig))
            errdata = msg
        elif nocc == 1 and flatocc is True:
            out = out[0]
            isempty = isempty[0]
    return {'data': out, 'units': unit,
            'isempty': isempty, 'errdata': errdata, 'errunits': errunits}


def get_data_units(ids=None, sig=None, occ=None,
                   data=None, units=None,
                   indch=None, indt=None, stack=None,
                   isclose=None, flatocc=True,
                   nan=True, pos=None, empty=None, strict=None, warn=True,
                   dids=None, dshort=None, dcomp=None, dall_except=None):
    """ Return a dict with the data and units (and empty, errors)

    For multiple shorcuts from the same ids

    For the desired ids and signal (shortcut), load the data and units
    Return a dict also containing a bool indicating whether the data is empty
    and error strings if any was raised

    """

    # ------------------
    # Check format input

    if strict is None:
        strict = _STRICT

    # ids = valid self.dids.keys()
    ids = _checkformat_getdata_ids(ids, dids=dids)

    # sig = list of str (shortcuts)
    sig, comp = _checkformat_getdata_sig(sig, ids,
                                         dshort=dshort, dcomp=dcomp,
                                         dall_except=dall_except)

    # occ = np.ndarray of valid int
    occ = _checkformat_getdata_occ(occ, ids, dids=dids)
    indoc = np.where(dids[ids]['occ'] == occ)[0]

    # Check all occ have isget = True
    indok = dids[ids]['isget'][indoc]
    if not np.all(indok):
        msg = ("All desired occurences shall have been gotten !\n"
               + "    - desired occ:   {}\n".format(occ)
               + "    - available occ: {}\n".format(dids[ids]['occ'])
               + "    - isget:         {}".format(dids[ids]['isget'])
               + "\n  => Try running self.open_get_close()")
        raise Exception(msg)

    # check indch if ids has channels
    if hasattr(dids[ids]['ids'][indoc[0]], 'channel'):
        nch = len(getattr(dids[ids]['ids'][indoc[0]], 'channel'))
        indch = _checkformat_getdata_indch(indch, nch)

    # ------------------
    # get data

    dout, dfail = {}, {}
    for ii in range(0, len(sig)):
        if isclose is None:
            isclose_ = sig[ii] == 't'
        else:
            isclose_ = isclose
        try:
            dout[sig[ii]] = _get_data_units(
                ids, sig[ii], occ, comp=bool(comp[ii]),
                indt=indt, indch=indch,
                stack=stack, isclose=isclose_, flatocc=flatocc,
                data=data, units=units,
                nan=nan, pos=pos, empty=empty, warn=warn,
                dids=dids, dcomp=dcomp, dshort=dshort,
                dall_except=dall_except)
            if dout[sig[ii]]['errdata'] is not None:
                dfail[sig[ii]] = dout[sig[ii]]['errdata']
                if warn is True:
                    msg = ('\n{}\n\t '.format(dout[sig[ii]]['errdata'])
                           + 'fail {0}.{1} data'.format(ids, sig[ii]))
                    warnings.warn(msg)
            if dout[sig[ii]]['errunits'] is not None:
                if warn is True:
                    msg = ('\n{}\n\t '.format(dout[sig[ii]]['errunits'])
                           + 'fail {0}.{1} units'.format(ids, sig[ii]))
                    warnings.warn(msg)

            # Remove if strict
            if strict is True:
                if dout[sig[ii]]['errdata'] is not None:
                    del dout[sig[ii]]

        except Exception as err:
            dfail[sig[ii]] = str(err)
            if warn is True:
                msg = ('\n{}\n\t '.format(dfail[sig[ii]])
                       + 'signal {0}.{1} not loaded!'.format(ids, sig[ii]))
                warnings.warn(msg)
    return dout, dfail



# -*- coding: utf-8 -*-
#! /usr/bin/python
"""
The imas-compatibility module of tofu

"""
import warnings
import traceback
import itertools as itt

try:
    try:
        from tofu.imas2tofu._core import *
        from tofu.imas2tofu._mat2ids2calc import *
    except Exception:
        from ._core import *
        from ._mat2ids2calc import *
except Exception as err:
    if str(err) == 'imas not available':
        msg = ""
        msg += "\n\nIMAS python API issue\n"
        msg += "imas could not be imported into tofu ('import imas' failed):\n"
        msg += "  - it may not be installed (optional dependency)\n"
        msg += "  - or you have loaded the wrong working environment\n\n"
        msg += "    => the optional sub-package tofu.imas2tofu is not usable\n"
    else:
        msg = str(traceback.format_exc())
        msg += "\n\n    => the optional sub-package tofu.imas2tofu is not usable\n"
    raise Exception(msg)


# -----------------------------------------------
#   Check IMAS version vs latest available in linux modules
# -----------------------------------------------

_KEYSTR = 'IMAS/'


# extract all IMAS versions from a str returned by modules
def extractIMAS(ss, keystr=_KEYSTR):
    if keystr not in ss:
        raise Exception
    ls = ss[ss.index(keystr):].split('\n')
    ls = itt.chain.from_iterable([s.split(' ') for s in ls])
    ls = [s for s in ls if keystr in s]
    ls = [s[len(keystr):s.index('(')] if '(' in s else s[len(keystr):]
          for s in ls]
    return sorted(ls)


# Compare current and latest available IMAS versions
def check_IMAS_version(verb=True, keystr=_KEYSTR):
    import subprocess

    # Get currently loaded IMAS
    cmd = "module list"
    proc = subprocess.run(cmd, check=True, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    lcur = extractIMAS(proc.stdout.decode(), keystr=keystr)
    if len(lcur) != 1:
        msg = ("You seem to have no / several IMAS version loaded:\n"
               + "\t- module list: {}".format(lcur))
        raise Exception(msg)

    # Get all available IMAS
    cmd = "module av IMAS"
    proc = subprocess.run(cmd, check=True, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    lav = extractIMAS(proc.stdout.decode(), keystr=keystr)
    if len(lav) == 0:
        msg = "There is not available IMAS version"
        raise Exception(msg)

    # Compare and warn
    if lcur[0] not in lav:
        msg = "The current IMAS version is not available!"
        raise Exception(msg)

    msg = None
    if lav.index(lcur[0]) != len(lav)-1:
        msg = ("\nYou do not seem to be using the latest IMAS version:\n"
               + "'module list' vs 'module av IMAS' suggests:\n"
               + "\t- Current version: {}\n".format(lcur[0])
               + "\t- Latest version : {}".format(lav[-1]))
        warnings.warn(msg)
    return lcur[0], lav


# Try comparing and warning
try:
    _, _ = check_IMAS_version(verb=True)
except Exception as err:
    # This warning is an optional luxury, should not block anything
    pass


__all__ = ['MultiIDSLoader', 'load_Config', 'load_Plasma2D',
           'load_Cam', 'load_Data']
del warnings, traceback, itt, _KEYSTR




# Built-in
import os
import re
import itertools as itt

# Common
import numpy as np
from scipy.interpolate import RectBivariateSpline as scpRectSpl


__all__ = ['read', 'read_all']


_DTYPES = {'adf11': ['acd', 'ccd', 'scd', 'plt', 'prb'],
           'adf15': None}
_DEG = 1


# #############################################################################
#                       Utility functions
# #############################################################################


def _get_PATH_LOCAL():
    pfe = os.path.join(os.path.expanduser('~'), '.tofu', 'openadas2tofu')
    if os.path.isdir(pfe):
        return pfe
    else:
        return None


def _get_subdir_from_pattern(path, pattern):
    ld = [dd for dd in os.listdir(path)
          if (os.path.isdir(os.path.join(path, dd))
              and pattern in dd)]
    if len(ld) != 1:
        av = [dd for dd in os.listdir(path)
              if os.path.isdir(os.path.join(path, dd))]
        msg = ("You have no / many directories in your local "
               + "~/.tofu/openadas2tofu/ matching the desired file type:\n"
               + "\t- provided : {}\n".format(pattern)
               + "\t- available: {}\n".format(av)
               + "  => download the data with tf.openadas2tofu.download()")
        raise Exception(msg)
    return os.path.join(path, ld[0])


# #############################################################################
#                       Main functions
# #############################################################################


def read(adas_path, **kwdargs):
    """ Read openadas-formatted files and return a dict with the data

    Povide the full adas file name
    The result is returned as a dict

    example
    -------
        >>> import tofu as tf
        >>> fn = '/adf11/scd74/scd74_ar.dat'
        >>> out = tf.openadas2tofu.read(fn)
        >>> fn = '/adf15/pec40][ar/pec40][ar_ca][ar16.dat'
        >>> out = tf.openadas2tofu.read(fn)
    """

    path_local = _get_PATH_LOCAL()

    # Check whether the local .tofu repo exists, if not recommend tofu-custom
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), '.tofu', 'openadas2tofu')
        msg = ("You do not seem to have a local ./tofu repository\n"
               + "tofu uses that local repository to store all user-specific "
               + "data and downloads\n"
               + "In particular, openadas files are downloaded and saved in:\n"
               + "\t{}\n".format(path)
               + "  => to set-up your local .tofu repo, run in a terminal:\n"
               + "\ttofu-custom")
        raise Exception(msg)

    # Determine whether adas_path is an absolute path or an adas full name
    if os.path.isfile(adas_path):
        pfe = adas_path
    else:
        # make sure adas_path is not understood as absolute local path
        if adas_path[0] == '/':
            adas_path = adas_path[1:]

        # Check file was downloaded locally
        pfe = os.path.join(path_local, adas_path)
        if not os.path.isfile(pfe):
            msg = ("Provided file does not seem to exist:\n"
                   + "\t{}\n".format(pfe)
                   + "  => Search it online with tofu.openadas2tofu.search()\n"
                   + "  => Download it with tofu.openadas2tofu.download()")
            raise FileNotFoundError(msg)

    lc = [ss for ss in _DTYPES.keys() if ss in pfe]
    if not len(lc) == 1:
        msg = ("File type could not be derived from absolute path:\n"
               + "\t- provided:  {}\n".format(pfe)
               + "\t- supported: {}".format(sorted(_DTYPES.keys())))
        raise Exception(msg)

    func = eval('_read_{}'.format(lc[0]))
    return func(pfe, **kwdargs)


def read_all(element=None, charge=None, typ1=None, typ2=None,
             verb=None, **kwdargs):
    """ Read all relevant openadas files for chosen typ1

    Please specify:
        - typ1: 'adf11' or 'adf15'
        - element: the symbol of the element

    If typ1 = 'adf11', you can also provide typ2 to specify the coefficients:
        - 'scd': effective ionisation coefficients
        - 'acd': effective electron-impact recombination coefficients
        - 'ccd': effective hydrogen-impact recombination coefficients
        - 'plt': line power due to electron-impact excitation
        - 'prc': line power due to hydrogen-impact excitation
        - 'prb': rad. recombination and bremmstrahlung due to electron-impact

    If typ1 = 'adf15', you can optioanlly provide a min/max wavelength

    The result is returned as a dict

    examples
    --------
        >>> import tofu as tf
        >>> dout = tf.openadas2tofu.read_all(element='ar', typ1='adf11')
        >>> dout = tf.openadas2tofu.read_all(element='ar', typ1='adf15',
                                             charge=16,
                                             lambmin=3.94e-10, lambmax=4.e-10)
    """

    # --------------------
    # Check whether the local .tofu repo exists, if not recommend tofu-custom
    path_local = _get_PATH_LOCAL()
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), '.tofu', 'openadas2tofu')
        msg = ("You do not seem to have a local ./tofu repository\n"
               + "tofu uses that local repository to store all user-specific "
               + "data and downloads\n"
               + "In particular, openadas files are downloaded and saved in:\n"
               + "\t{}\n".format(path)
               + "  => to set-up your local .tofu repo, run in a terminal:\n"
               + "\ttofu-custom")
        raise Exception(msg)

    # --------------------
    # Check / format input
    if typ1 is None:
        typ1 = 'adf15'
    if not isinstance(typ1, str) or typ1.lower() not in _DTYPES.keys():
        msg = ("Please choose a valid adas file type:\n"
               + "\t- allowed:  {}\n".format(_DTYPES.keys())
               + "\t- provided: {}".format(typ1))
        raise Exception(msg)
    typ1 = typ1.lower()
    if typ1 == 'adf11' and typ2 is None:
        typ2 = _DTYPES[typ1]
        fd = os.listdir(os.path.join(path_local, typ1))
        typ2 = [sorted([ss for ss in fd if tt in ss])[-1] for tt in typ2]
    if isinstance(typ2, str):
        typ2 = [typ2]
    if (_DTYPES[typ1] is not None
        and (not isinstance(typ2, list)
             or not all([any([s1 in ss for s1 in _DTYPES[typ1]])
                         for ss in typ2]))):
        msg = ("typ2 must be a list of valid openadas file types for typ1:\n"
               + "\t- provided:         {}\n".format(typ2)
               + "\t- available for {}: {}".format(typ1, _DTYPES[typ1]))
        raise Exception(msg)

    if not isinstance(element, str):
        msg = "Please choose an element!"
        raise Exception(msg)
    element = element.lower()
    if charge is not None and not isinstance(charge, int):
        msg = "charge must be a int!"
        raise Exception(msg)

    if verb is None:
        verb = True

    # --------------------
    # Get list of relevant directories

    # Level 1: Type
    path = _get_subdir_from_pattern(path_local, typ1)
    # Level 2: element or typ2
    if typ1 == 'adf11':
        lpath = [_get_subdir_from_pattern(path, tt) for tt in typ2]
    elif typ1 == 'adf15':
        lpath = [_get_subdir_from_pattern(path, element)]

    # --------------------
    # Get list of relevant files pfe
    lpfe = list(itt.chain.from_iterable(
        [[os.path.join(path, ff) for ff in os.listdir(path)
          if (os.path.isfile(os.path.join(path, ff))
              and ff[-4:] == '.dat'
              and element in ff)]
         for path in lpath]))

    if charge is not None and typ1 == 'adf15':
        lpfe = [ff for ff in lpfe if str(charge) in ff]

    # --------------------
    # Extract data from each file
    func = eval('_read_{}'.format(typ1))
    dout = {}
    for pfe in lpfe:
        if verb is True:
            msg = "\tLoading data from {}".format(pfe)
            print(msg)
        out = func(pfe, dout=dout, **kwdargs)
    return out


# #############################################################################
#                      Specialized functions for ADF 11
# #############################################################################


def _read_adf11(pfe, deg=None, dout=None):
    if deg is None:
        deg = _DEG
    if dout is None:
        dout = {}

    # Get second order file type
    typ1 = [vv for vv in _DTYPES['adf11'] if vv in pfe]
    if len(typ1) != 1:
        msg = ("Second order file type culd not be inferred from file name!\n"
               + "\t- available: {}\n".format(_DTYPES['adf11'])
               + "\t- provided: {}".format(pfe))
        raise Exception(msg)
    typ1 = typ1[0]

    # Get element
    elem = pfe[:-4].split('_')[1]
    comline = '-'*60
    comline2 = 'C'+comline

    if typ1 in ['acd', 'ccd', 'scd', 'plt', 'prb']:

        # read blocks
        with open(pfe) as search:
            for ii, line in enumerate(search):

                if comline2 in line:
                    break

                # Get atomic number (transitions) stored in this file
                if ii == 0:
                    lstr = line.split('/')
                    lin = [ss for ss in lstr[0].strip().split(' ')
                           if ss.strip() != '']
                    lc = [len(lin) == 5 and all([ss.isdigit() for ss in lin]),
                          elem.upper() in lstr[1],
                          'ADF11' in lstr[2]]
                    if not all(lc):
                        msg = ("File header format seems to have changed!\n"
                               + "\t- lc = {}".format(lc))
                        raise Exception(msg)
                    Z, nne, nte, q0, qend = map(int, lin)
                    nelog10 = np.array([])
                    telog10 = np.array([])
                    in_ne = True
                    continue

                if comline in line:
                    continue

                # Get nelog10
                if in_ne:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    nelog10 = np.append(nelog10, np.array(li, dtype=float))
                    if nelog10.size == nne:
                        in_ne = False
                        in_te = True

                # Get telog10
                elif in_te is True:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    telog10 = np.append(telog10, np.array(li, dtype=float))
                    if telog10.size == nte:
                        in_te = False
                        in_ion = True

                # Get ion block
                elif (in_ion is True and 'Z1=' in line
                      and ('------/' in line and 'DATE=' in line)):
                    nion = int(
                        line[line.index('Z1=')+len('Z1='):].split('/')[0])
                    if typ1 in ['scd', 'plt']:
                        charge = nion - 1
                    else:
                        charge = nion
                    coefslog10 = np.array([])
                elif in_ion is True and charge is not None:
                    li = [ss for ss in line.strip().split(' ')
                          if ss.strip() != '']
                    coefslog10 = np.append(coefslog10,
                                           np.array(li, dtype=float))
                    if coefslog10.size == nne*nte:
                        key = '{}{}'.format(elem, charge)
                        tkv = [('element', elem), ('Z', Z), ('charge', charge)]
                        if key in dout.keys():
                            assert all([dout[key][ss] == vv for ss, vv in tkv])
                        else:
                            dout[key] = {ss: vv for ss, vv in tkv}
                        if typ1 == 'scd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['ionis'] = {'func': func,
                                                  'type': 'log10_nete',
                                                  'units': 'log10(m3/s)',
                                                  'source': pfe}
                        elif typ1 == 'acd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['recomb'] = {'func': func,
                                                   'type': 'log10_nete',
                                                   'units': 'log10(m3/s)',
                                                   'source': pfe}
                        elif typ1 == 'ccd':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10-6 to convert cm3/s -> m3/s
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))-6,
                                              kx=deg, ky=deg)
                            dout[key]['recomb_ce'] = {'func': func,
                                                      'type': 'log10_nete',
                                                      'units': 'log10(m3/s)',
                                                      'source': pfe}
                        elif typ1 == 'plt':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10+6 to convert W.cm3 -> W.m3
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))+6,
                                              kx=deg, ky=deg)
                            dout[key]['rad_bb'] = {'func': func,
                                                   'type': 'log10_nete',
                                                   'units': 'log10(W.m3)',
                                                   'source': pfe}
                        elif typ1 == 'prb':
                            # nelog10+6 to convert /cm3 -> /m3
                            # coefslog10+6 to convert W.cm3 -> W.m3
                            func = scpRectSpl(nelog10+6, telog10,
                                              coefslog10.reshape((nne, nte))+6,
                                              kx=deg, ky=deg)
                            dout[key]['rad_fffb'] = {'func': func,
                                                     'type': 'log10_nete',
                                                     'units': 'log10(W.m3)',
                                                     'source': pfe}
                        if nion == Z:
                            break

    return dout


# #############################################################################
#                      Specialized functions for ADF 15
# #############################################################################


def _get_adf15_key(elem, charge, isoel, typ0, typ1):
    return '{}{}_{}_openadas_{}_{}'.format(elem, charge, isoel,
                                           typ0, typ1)


def _read_adf15(pfe, dout=None,
                lambmin=None,
                lambmax=None,
                deg=None):

    if deg is None:
        deg = _DEG
    if dout is None:
        dout = {}

    # Get summary of transitions
    flagblock = '/isel ='
    flag0 = 'superstage partition information'

    # Get file markers from name (elem, charge, typ0, typ1)
    typ0, typ1, elemq = pfe.split('][')[1:]
    ind = re.search(r'\d', elemq).start()
    elem = elemq[:ind].title()
    charge = int(elemq[ind:-4])
    assert elem.lower() in typ0[:2]
    assert elem.lower() == typ1.split('_')[0]
    typ0 = typ0[len(elem)+1:]
    typ1 = typ1.split('_')[1]

    # Extract data from file
    nlines, nblock = None, 0
    in_ne, in_te, in_pec, in_tab, itab = False, False, False, False, np.inf
    skip = False
    with open(pfe) as search:
        for ii, line in enumerate(search):

            # Get number of lines (transitions) stored in this file
            if ii == 0:
                lstr = line.split('/')
                nlines = int(lstr[0].replace(' ', ''))
                continue

            # Get info about the transition being scanned (block)
            if flagblock in line and 'C' not in line and nblock < nlines:
                lstr = [kk for kk in line.rstrip().split(' ') if len(kk) > 0]
                lamb = float(lstr[0])*1.e-10
                isoel = nblock + 1
                nblock += 1
                c0 = ((lambmin is not None and lamb < lambmin)
                      or (lambmax is not None and lamb > lambmax))
                if c0:
                    skip = True
                    continue
                skip = False
                nne, nte = int(lstr[1]), int(lstr[2])
                typ = [ss[ss.index('type=')+len('type='):ss.index('/ispb')]
                       for ss in lstr[3:] if 'type=' in ss]
                assert len(typ) == 1
                # To be updated : proper rezading from line
                in_ne = True
                ne = np.array([])
                te = np.array([])
                pec = np.full((nne*nte,), np.nan)
                ind = 0
                continue

            if 'root partition information' in line and skip is True:
                skip = False

            # Check lamb is ok
            if skip is True:
                continue

            # Get ne for the transition being scanned (block)
            if in_ne is True:
                ne = np.append(ne,
                               np.array(line.rstrip().strip().split(' '),
                                        dtype=float))
                if ne.size == nne:
                    in_ne = False
                    in_te = True

            # Get te for the transition being scanned (block)
            elif in_te is True:
                te = np.append(te,
                               np.array(line.rstrip().strip().split(' '),
                                        dtype=float))
                if te.size == nte:
                    in_te = False
                    in_pec = True

            # Get pec for the transition being scanned (block)
            elif in_pec is True:
                data = np.array(line.rstrip().strip().split(' '),
                                dtype=float)
                pec[ind:ind+data.size] = data
                ind += data.size
                if ind == pec.size:
                    in_pec = False
                    key = _get_adf15_key(elem, charge, isoel, typ0, typ1)
                    # log(ne)+6 to convert /cm3 -> /m3
                    # log(pec)+6 to convert cm3/s -> m3/s
                    func = scpRectSpl(np.log(ne)+6, np.log(te),
                                      np.log(pec).reshape((nne, nte))+6,
                                      kx=deg, ky=deg)
                    dout[key] = {'lambda': lamb,
                                 'ION': '{}{}+'.format(elem, charge),
                                 'symbol': '{}{}-{}'.format(typ0, typ1, isoel),
                                 'origin': pfe,
                                 'type': typ[0],
                                 'ne': ne, 'te': te,
                                 'pec': {'func': func,
                                         'type': 'log_nete',
                                         'units': 'log(m3/s)',
                                         'source': 'pfe'}}

            # Get transitions from table at the end
            if 'photon emissivity atomic transitions' in line:
                itab = ii + 6
            if ii == itab:
                in_tab = True
            if in_tab is True:
                lstr = [kk for kk in line.rstrip().split(' ') if len(kk) > 0]
                isoel = int(lstr[1])
                lamb = float(lstr[2])*1.e-10
                key = _get_adf15_key(elem, charge, isoel, typ0, typ1)
                c0 = ((lambmin is None or lambmin < lamb)
                      and (lambmax is None or lambmax > lamb))
                if c0 and key not in dout.keys():
                    msg = ("Inconsistency in file {}:\n".format(pfe)
                           + "\t- line should be present".format(key))
                    raise Exception(msg)
                if key in dout.keys():
                    if dout[key]['lambda'] != lamb:
                        msg = "Inconsistency in file {}".format(pfe)
                        raise Exception(msg)
                    c0 = (dout[key]['type'] not in lstr
                          or lstr.index(dout[key]['type']) < 4)
                    if c0:
                        msg = ("Inconsistency in table, type not found:\n"
                               + "\t- expected: {}\n".format(dout[key]['type'])
                               + "\t- line: {}".format(line))
                        raise Exception(msg)
                    trans = lstr[3:lstr.index(dout[key]['type'])]
                    dout[key]['transition'] = ''.join(trans)
                if isoel == nlines:
                    in_tab = False
    assert all(['transition' in vv.keys() for vv in dout.values()])
    return dout



# distutils: language=c++
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions for the earclipping techinque for triangulation of a
# polygon. This is really useful when vignetting, as the computation to know
# if a ray intersected a polygon, is easier if we check if the polygon is
# discretized in triangles and then we check if the ray intersected each
# triangle.
################################################################################
cdef  void compute_diff3d(double* orig,
                          int nvert,
                          double* diff) nogil

cdef  void are_points_reflex(int nvert,
                             double* diff,
                             bint* are_reflex) nogil

cdef  bint is_pt_in_tri(double[3] v0, double[3] v1,
                        double ax, double ay, double az,
                        double px, double py, double pz) nogil

cdef  void earclipping_poly(double* vignett,
                            long* ltri,
                            double* diff,
                            bint* lref,
                            int nvert) nogil

cdef int triangulate_polys(double** vignett_poly,
                            long* lnvert,
                            int nvign,
                            long** ltri,
                            int num_threads) nogil except -1

cdef void vignetting_core(double[:, ::1] ray_orig,
                          double[:, ::1] ray_vdir,
                          double** vignett,
                          long* lnvert,
                          double* lbounds,
                          long** ltri,
                          int nvign,
                          int nlos,
                          bint* goes_through,
                          int num_threads) nogil



"""
This module stores all the default setting of ToFu
Including in particular computing parameters, dictionnaries and figures
"""

#import matplotlib
#matplotlib.use('WxAgg')
#matplotlib.interactive(True)

import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.path import Path
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

#from mayavi import mlab
import datetime as dtm
import time as time

# tofu
try:
    from tofu.version import __version__
    import tofu.utils as utils
except Exception:
    from tofu.version import __version__
    from .. import utils as utils


__github = 'https://github.com/ToFuProject/tofu/issues'
_wintit = 'tofu-%s        report issues / requests at %s'%(__version__, __github)
_fontsize = 8
_labelpad = 0


"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings generic
###############################################################################
###############################################################################
"""



"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_Geom
###############################################################################
###############################################################################
"""


# ---------- Common to several classes ----------

Legpropd = {'size':10}
TorLegd = {'frameon':False,'ncol':1,'bbox_to_anchor':(1.01, 1),'loc':2,'borderaxespad':0.,'prop':Legpropd}


#####################################################################
##################  Ves and Struct class  ###########################
#####################################################################


# ------------ Computing settings ---------------

TorNP = 50
TorRelOff = 0.05
TorInsideNP = 100
TorSplprms = [100.,2.,3]
DetBaryCylNP1 = 50
DetBaryCylNP2 = 200

# --- Plotting dictionaries and parameters ------

TorPd = {'c':'k','lw':2}
TorId = {'c':'k','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorBsd = {'c':'b','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorBvd = {'c':'g','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorVind = {'color':'r','scale':10}
TorITord = {'c':'k','ls':'dashed'}
TorBsTord = {'c':'b','ls':'dashed'}
TorBvTord = {'c':'g','ls':'dashed'}
TorNTheta = 50
Tor3DThetalim = [np.pi/2,2*np.pi]
Tor3DThetamin = np.pi/20.
TorP3Dd = {'color':(0.8,0.8,0.8,1.),'rstride':1,'cstride':1,'linewidth':0, 'antialiased':False}
TorPFilld = {'edgecolor':(0.8,0.8,0.8,1.),'facecolor':(0.8,0.8,0.8,1.),'linestyle':'solid','linewidth':1}
TorPAng = 'theta'
TorPAngUnit = 'rad'
TorPSketch = True
TorP3DFilld = {'color':(0.8,0.8,0.8,0.4),'linestyle':'solid','linewidth':0}

Vesdict = dict(Lax=None, Proj='All', Elt='PIBsBvV', dP=TorPd, dI=TorId,
               dBs=TorBsd, dBv=TorBvd, dVect=TorVind, dIHor=TorITord,
               dBsHor=TorBsTord, dBvHor=TorBvTord, dLeg=TorLegd,
               Lim=Tor3DThetalim, Nstep=TorNTheta, draw=True, Test=True)


StructPd = {'fc':(0.8,0.8,0.8,0.5),'ec':'k','linewidth':1}
StructPd_Tor = {'fc':(0.8,0.8,0.8,0.5),'ec':'none'}

Structdict = dict(Vesdict)
Structdict['dP'] = StructPd_Tor
#Structdict['P3Dd'] = {'color':(0.8,0.8,0.8,1.),'rstride':1,'cstride':1,
#                      'linewidth':0, 'antialiased':False}


# -------------- Figures ------------------------


def Plot_LOSProj_DefAxes(Mode, Type='Tor', fs=None, wintit='tofu'):
    assert fs is None or (type(fs) is str and fs=='a4') or len(fs)==2
    assert Mode.lower() in ['cross','hor','all'], "Arg should be 'Cross' or 'Hor' or 'All' !"
    assert Type in ['Tor','Lin'], "Arg Type must be in ['Tor','Lin'] !"
    assert wintit is None or type(wintit) is str, "Arg wintit must be a str !"
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    if wintit is not None:
        f.canvas.set_window_title(wintit)
    if Mode.lower()=='all':
        axPosP, axPosT = [0.07, 0.1, 0.3, 0.8], [0.55, 0.1, 0.3, 0.8]
        axP = f.add_axes(axPosP,frameon=True,facecolor=axCol)
        axT = f.add_axes(axPosT,frameon=True,facecolor=axCol)
        if Type=='Tor':
            axP.set_xlabel(r"R (m)"),   axP.set_ylabel(r"Z (m)")
        else:
            axP.set_xlabel(r"Y (m)"),   axP.set_ylabel(r"Z (m)")
        axT.set_xlabel(r"X (m)"),   axT.set_ylabel(r"Y (m)")
        axP.set_aspect(aspect="equal", adjustable='datalim')
        axT.set_aspect(aspect="equal", adjustable='datalim')
        ax = [axP,axT]
    else:
        axPos = [0.15, 0.15, 0.6, 0.7]
        ax = f.add_axes(axPos,frameon=True,facecolor=axCol)
        if Mode == 'Cross':
            if Type=='Tor':
                ax.set_xlabel(r"R (m)"),    ax.set_ylabel(r"Z (m)")
            else:
                ax.set_xlabel(r"Y (m)"),    ax.set_ylabel(r"Z (m)")
        elif Mode == 'Hor':
            ax.set_xlabel(r"X (m)"),    ax.set_ylabel(r"Y (m)")
        ax.set_aspect(aspect="equal", adjustable='datalim')
    return ax

def Plot_3D_plt_Tor_DefAxes(fs=None, wintit='tofu'):
    assert fs is None or (type(fs) is str and fs=='a4') or len(fs)==2
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (14,10)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    if wintit is not None:
        f.canvas.set_window_title(wintit)
    axPos = [0.05, 0.05, 0.75, 0.85]
    ax = f.add_axes(axPos,facecolor=axCol,projection='3d')
    ax.set_xlabel(r"X (m)")
    ax.set_ylabel(r"Y (m)")
    ax.set_zlabel(r"Z (m)")
    # ax.set_aspect(aspect="equal", adjustable='datalim')
    return ax


def Plot_Impact_DefAxes(Proj, Ang='theta', AngUnit='rad', fs=None, wintit='tofu', Sketch=True):
    assert fs is None or (type(fs) is str and fs=='a4') or len(fs)==2
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (10,6) if Proj.lower()=='Cross' else (11,9)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    if wintit is not None:
        f.canvas.set_window_title(wintit)
    if Proj == 'Cross':
        axPos = [0.12, 0.12, 0.60, 0.8]
        ax, axSketch = f.add_axes(axPos,frameon=True,facecolor=axCol), []
        XAng = r"$\theta$" if Ang=='theta' else r"$\xi$"
        XUnit = r"$(rad.)$" if AngUnit=='rad' else r"$(deg.)$"
        XTickLab = [r"$0$",r"$\pi/4$",r"$\pi/2$",r"$3\pi/4$",r"$\pi$"] if AngUnit=='rad' else [r"$0$",r"$90$",r"$180$",r"$270$",r"$360$"]
        ax.set_xlabel(XAng+r" "+XUnit)
        ax.set_ylabel(r"$p$ $(m)$")
        ax.set_xlim(0,np.pi)
        ax.set_ylim(-1.5,1.5)
        ax.set_xticks(np.pi*np.array([0.,1./4.,1./2.,3./4.,1.]))
        ax.set_xticklabels(XTickLab)
        if Sketch:
            axSketch = f.add_axes([0.75, 0.10, 0.15, 0.15],frameon=False,facecolor=axCol)
            Pt, Line, Hor, theta, ksi = np.array([[0,-0.8],[0,0.8]]), np.array([[-1.6,0.1],[0,1.7]]), np.array([[-0.4,0.2],[1.2,1.2]]), np.linspace(0,3.*np.pi/4.,30), np.linspace(0,np.pi/4.,10)
            theta, ksi = np.array([0.3*np.cos(theta),0.3*np.sin(theta)]), np.array([-0.4+0.4*np.cos(ksi), 1.2+0.4*np.sin(ksi)])
            axSketch.plot(Pt[0,:],Pt[1,:],'+k',Pt[0,:],Pt[1,:],'--k',Line[0,:],Line[1,:],'-k', Hor[0,:],Hor[1,:],'-k', theta[0,:],theta[1,:],'-k', ksi[0,:],ksi[1,:],'-k')
            axSketch.annotate(r"$\theta$", xy=(0.3,0.4),xycoords='data',va="center", ha="center")
            axSketch.annotate(r"$\xi$", xy=(0.1,1.4),xycoords='data',va="center", ha="center")
            axSketch.annotate(r"$p$", xy=(-0.7,0.3),xycoords='data',va="center", ha="center")
            axSketch.set_xticks([]), axSketch.set_yticks([])
            axSketch.axis("equal")
        return ax, axSketch
    elif Proj.lower() == '3d':
        axPos = [0.1, 0.1, 0.65, 0.8]
        ax = f.add_axes(axPos,facecolor=axCol,projection='3d')
        XAng = r"$\theta$" if Ang=='theta' else r"$\xi$"
        XUnit = r"$(rad.)$" if AngUnit=='rad' else r"$(deg.)$"
        XTickLab = [r"$0$",r"$\pi/4$",r"$\pi/2$",r"$3\pi/4$",r"$\pi$"] if AngUnit=='rad' else [r"$0$",r"$90$",r"$180$",r"$270$",r"$360$"]
        ax.set_xlabel(XAng+r" "+XUnit)
        ax.set_ylabel(r"$p$ $(m)$")
        ax.set_zlabel(r"$\phi$ $(rad)$")
        ax.set_xlim(0,np.pi)
        ax.set_ylim(-1.5,1.5)
        ax.set_zlim(-np.pi/2.,np.pi/2.)
        ax.set_xticks(np.pi*np.array([0.,1./4.,1./2.,3./4.,1.]))
        ax.set_xticklabels(XTickLab)
        return [ax]

#def Plot_3D_mlab_Tor_DefFig():
#    fW,fH,fBgC = 700,500,(1.,1.,1.)
#    axPosP, axPosT = [0.07, 0.1, 0.3, 0.8], [0.55, 0.1, 0.3, 0.8]
#    f = mlab.figure(bgcolor=fBgC,fgcolor=None,size=(fW,fH))
#    return f



def _Config_phithetaproj_default(fs=None, dmargin=None,
                                 fontsize=_fontsize, labelpad=_labelpad,
                                 wintit=_wintit, ntMax=1):

    fldict = dict(fontsize=fontsize, labelpad=labelpad)

    # Figure
    axCol = "w"
    fs = utils.get_figuresize(fs, fsdef=(12,7))
    if dmargin is None:
        dmargin = dict(left=0.06, bottom=0.06, right=0.97, top=0.95,
                       wspace=0.8, hspace=0.3)
    fig = plt.figure(facecolor=axCol,figsize=fs)
    if wintit is not None:
        fig.canvas.set_window_title(wintit)

    # -------------
    # Axes grid
    # -------

    gs0 = gridspec.GridSpec(5, 2, **dmargin)

    axt = fig.add_subplot(gs0[0,:], fc='w')
    axd = fig.add_subplot(gs0[1:3,:], fc='w')
    axc = fig.add_subplot(gs0[3:,0], fc='w')
    axh = fig.add_subplot(gs0[3:,1], fc='w')

    axc.set_aspect('equal', adjustable='datalim')
    axh.set_aspect('equal', adjustable='datalim')

    axt.set_xlabel(r'$s$ (s)', **fldict)
    axt.set_ylabel(r'$data$ (a.u.)', **fldict)
    axd.set_xlabel(r'$\phi$ (rad)', **fldict)
    axd.set_ylabel(r'$\theta$ (rad)', **fldict)
    axc.set_xlabel(r'$R$ (m)', **fldict)
    axc.set_ylabel(r'$Z$ (m)', **fldict)
    axh.set_xlabel(r'$X$ (m)', **fldict)
    axh.set_ylabel(r'$Y$ (m)', **fldict)

    dax = {'t':[axt],
           'dist':[axd],
           'cross':[axc],
           'hor':[axh]}

    return fig, dax





#####################################################################
########################  LOS class  ################################
#####################################################################


# ------------ Computing settings ---------------
LOSDiscrtSLim = (0.,1,)
LOSDiscrtSLMode = 'norm'
LOSDiscrtDS = 0.005
LOSDiscrtSMode = 'm'


# --- Plotting dictionaries and parameters ------

LOSLd = {'c':'k','lw':2}
LOSMd = {'c':'k','ls':'None','lw':2,'marker':'x','markersize':8,'mew':2}
LOSMImpd = {'c':'k','ls':'None','lw':2,'marker':'x','markersize':8,'mew':2}
LOSLplot = 'Tot'
LOSImpAng = 'theta'
LOSImpAngUnit = 'rad'
LOSImpElt = 'LV'

LOSdict = dict(Lax=None, Proj='All', Lplot=LOSLplot, Elt='LDIORP', EltVes='', Leg='', Ldict=LOSLd, MdictD=LOSMd, MdictI=LOSMd, MdictO=LOSMd, MdictR=LOSMd, MdictP=LOSMd, LegDict=TorLegd, Vesdict=Vesdict, draw=True, Test=True)

# -------------- Figures ------------------------






"""
#####################################################################
###################### Lens class  ################################
#####################################################################






# -------------- Figures ------------------------

def Plot_Lens_Alone_DefAxes(fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (20,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    axPos = [0.05, 0.1, 0.9, 0.85]
    ax = f.add_axes(axPos,facecolor=axCol)
    ax.set_xlabel(r"x (m)")
    ax.set_ylabel(r"y (m)")
    return ax




#####################################################################
###################### Detect class  ################################
#####################################################################


# ------------ Computing settings ---------------

DetSpanRMinMargin = 0.9
DetSpanNEdge = 5
DetSpanNRad = 5

DetConeNEdge = 8
DetConeNRad = 6

DetPreConedX12 = [0.01, 0.01]
DetPreConedX12Mode = 'abs'
DetPreConeds = 0.01
DetPreConedsMode = 'abs'
DetPreConeMarginS = 0.002

DetConeDX = 0.002
DetConeDRY = 0.0025 # 0.0025
DetConeDTheta = np.pi/1024. # 512.
DetConeDZ = 0.0025 # 0.0025
#DetConeNTheta = 25 # 25
#DetConeNZ = 50 # 50

DetConeRefdMax = 0.02

DetEtendMethod = 'quad'
DetEtenddX12 = [0.01, 0.01]
DetEtenddX12Mode = 'rel'
DetEtendepsrel = 1.e-3
DetEtendRatio = 0.02
DetCalcEtendColis = False

DetCalcSAngVectColis = True
DetCalcSAngVectCone = True

DetSynthEpsrel = 1.e-4
DetSynthdX12 = [0.005, 0.005]
DetSynthdX12Mode = 'abs'
DetSynthds = 0.005
DetSynthdsMode = 'abs'
DetSynthMarginS = 0.001

# --- Plotting dictionaries and parameters ------

ApPd = {'c':'k','lw':2,'ls':'solid'}
ApVd = {'color':'r','lw':2,'ls':'solid'}
DetPd = {'c':'k','lw':2,'ls':'solid'}
DetVd = {'color':'r','lw':2,'ls':'solid'}
DetSAngPld = {'cmap':plt.cm.YlOrRd,'lw':0.,'rstride':1,'cstride':1, 'antialiased':False, 'edgecolor':'None'}
DetSangPlContd = {'linewidths':0.}
DetConed = {'edgecolors':'k', 'facecolors':(0.8,0.8,0.8,0.2), 'alpha':0.2, 'linewidths':0., 'linestyles':'-', 'antialiaseds':False}
DetImpd = {'ls':'solid','c':'k','lw':1}

ApLVin = 0.1
DetSAngPlRa = 0.5
DetSAngPldX12 = [0.025,0.025]
DetSAngPldX12Mode = 'rel'
DetSAngPlRatio = 0.01
DetEtendOnLOSNP = 20
DetEtendOnLOSModes = ['trapz']
DetEtendOnLOSLd = {'ls':'solid','c':'k','lw':2}

DetSAngPlot = 'Int'
DetSAngPlotMode = 'scatter'
DetSAngPlotd = {'cmap':plt.cm.YlOrRd}
DetSAngPlotLvl = 20

DetSliceAGdR = 0.005
DetSliceAGdY = 0.005
DetSliceAGdX = 0.01
DetSliceAGdTheta = np.pi/512.
DetSliceAGdZ = 0.005
DetSliceNbd = {'scatter':{'cmap':plt.cm.Greys,'marker':'s','edgecolors':'None','s':10},
        'contour':{'cmap':plt.cm.Greys},
        'contourf':{'cmap':plt.cm.Greys},
        'imshow':{'cmap':plt.cm.Greys}}
DetSliceSAd = {'scatter':{'cmap':plt.cm.YlOrRd,'marker':'s','edgecolors':'None','s':10, 'vmin':0},
        'contour':{'cmap':plt.cm.YlOrRd, 'vmin':0},
        'contourf':{'cmap':plt.cm.YlOrRd, 'vmin':0},
        'imshow':{'cmap':plt.cm.YlOrRd, 'vmin':0}}


DetPolProjNTheta = 50
DetPolProjNZ = 25
DetSAngColis = True

GDetEtendMdA = {'ls':'None','c':'k','lw':2,'marker':'+'}
GDetEtendMdR = {'ls':'None','c':'b','lw':2,'marker':'x'}
GDetEtendMdS = {'ls':'None','c':'g','lw':2,'marker':'o'}
GDetEtendMdP = {'ls':'None','c':'r','lw':2,'marker':'*'}

GDetSigd = {'ls':'solid','c':'k','lw':2,'marker':'+'}


Apertdict = dict(Lax=None, Proj='All', Elt='PV', EltVes='', Leg='', LVIn=ApLVin, Pdict=ApPd, Vdict=ApVd, Vesdict=Vesdict, LegDict=TorLegd, draw=True, Test=True)
#Detdict =


# -------------- Figures ------------------------

def Plot_SAng_Plane_DefAxes(fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (10,8,80,'w')
    axPos = [0.05, 0.05, 0.9, 0.9]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,facecolor=axCol,projection='3d')
    ax.set_xlabel(r"X1 (m)")
    ax.set_ylabel(r"X2 (m)")
    ax.set_zlabel(r"$\Omega$ (sr)")
    return ax


def Plot_Etendue_AlongLOS_DefAxes(kMode='rel',fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (14,8,80,'w')
    axPos = [0.06, 0.08, 0.70, 0.86]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,facecolor=axCol)
    if kMode.lower()=='rel':
        ax.set_xlabel(r"Rel. length (adim.)")
    else:
        ax.set_xlabel(r"Length (m)")
    ax.set_ylabel(r"Etendue ($sr.m^2$)")
    return ax


def Plot_CrossSlice_SAngNb_DefAxes(VType='Tor', fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (15,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    axSAng = f.add_axes([0.05, 0.06, 0.40, 0.85],frameon=True,facecolor=axCol)
    axNb = f.add_axes([0.60, 0.06, 0.40, 0.85],frameon=True,facecolor=axCol)
    if VType=='Tor':
        axSAng.set_xlabel(r"R (m)"),    axNb.set_xlabel(r"R (m)")
    elif VType=='Lin':
        axSAng.set_xlabel(r"Y (m)"),    axNb.set_xlabel(r"Y (m)")
    axSAng.set_ylabel(r"Z (m)"),      axNb.set_ylabel(r"Z (m)")
    axSAng.set_aspect(aspect="equal", adjustable='datalim')
    axNb.set_aspect(aspect="equal", adjustable='datalim')
    return axSAng, axNb


def Plot_HorSlice_SAngNb_DefAxes(fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (15,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    axSAng = f.add_axes([0.07, 0.12, 0.35, 0.8],frameon=True,facecolor=axCol)
    axNb = f.add_axes([0.55, 0.12, 0.35, 0.8],frameon=True,facecolor=axCol)
    axSAng.set_xlabel(r"X (m)"),    axSAng.set_ylabel(r"Y (m)")
    axNb.set_xlabel(r"X (m)"),      axNb.set_ylabel(r"Y (m)")
    axSAng.set_aspect(aspect="equal", adjustable='datalim')
    axNb.set_aspect(aspect="equal", adjustable='datalim')
    return axSAng, axNb


def Plot_Etendues_GDetect_DefAxes(fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.05,0.1,0.85,0.80],frameon=True,facecolor=axCol)
    ax.set_xlabel(r"")
    ax.set_ylabel(r"Etendue (sr.m^2)")
    return ax


def Plot_Sig_GDetect_DefAxes(fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.05,0.1,0.85,0.80],frameon=True,facecolor=axCol)
    ax.set_xlabel(r"")
    ax.set_ylabel(r"Signal (W)")
    return ax


#Ldict_mlab_Def = {'color':(0.,0.,0.),'tube_radius':None}
#Mdict_mlab_Def = {'color':(0.,0.,0.),'line_width':1,'mode':'sphere'}
#Dict_3D_mlab_Tor_Def = {'color':(0.8,0.8,0.8),'opacity':0.15,'transparent':False,'scale_factor':0.1}


def Plot_GDetect_Resolution_DefAxes(VType='Tor', fs=None, wintit='tofu'):
    axCol, fdpi = 'w', 80
    if fs is None:
        fs = (16,8) if Mode.lower()=='all' else (6,8)
    elif type(fs) is str and fs=='a4':
        fs = (11.69,8.27) if Mode.lower()=='all' else (8.27,11.69)
    f = plt.figure(facecolor="w", figsize=fs, dpi=fdpi)
    axCol = "w"
    (fW,fH) = (11.69,8.27) if a4 else (16,10)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol)
    ax1 = f.add_axes([0.05, 0.06, 0.32, 0.80], frameon=True, facecolor=axCol)
    ax2 = f.add_axes([0.50, 0.55, 0.47, 0.40], frameon=True, facecolor=axCol)
    ax3 = f.add_axes([0.50, 0.06, 0.47, 0.40], frameon=True, facecolor=axCol)
    X1 = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax1.set_xlabel(X1)
    ax1.set_ylabel(r"Z (m)")
    ax2.set_xlabel(r"size (a.u.)")
    ax2.set_ylabel(r"Signal (mW)")
    ax3.set_xlabel(r"Channels index (from 0)")
    ax3.set_ylabel(r"Signal (mW)")
    ax1.set_aspect(aspect='equal',adjustable='datalim')
    return ax1, ax2, ax3








"""




import numpy as np
import scipy.interpolate as scpinterp

# ###############################################
# ###############################################
#           CrystalBragg
# ###############################################
# ###############################################

# ###############################################
#           sampling
# ###############################################

def CrystBragg_sample_outline_sphrect(dpsi, dtheta, npsi=None, ntheta=None):
    psi = dpsi*np.linspace(-1, 1., npsi)
    theta = np.pi/2. + dtheta*np.linspace(-1, 1., ntheta)
    psimin = np.full((ntheta,), psi[0])
    psimax = np.full((ntheta,), psi[-1])
    thetamin = np.full((npsi,), theta[0])
    thetamax = np.full((npsi,), theta[-1])
    psi = np.concatenate((psi, psimax,
                          psi[::-1], psimin))
    theta = np.concatenate((thetamin, theta,
                            thetamax, theta[::-1]))
    return psi, theta

def CrystBragg_get_noute1e2_from_psitheta(nout, e1, e2, psi, theta,
                                          e1e2=True):
    vout = ((np.cos(psi)[None, :]*nout[:, None]
             + np.sin(psi)[None, :]*e1[:, None])*np.sin(theta)[None, :]
            + np.cos(theta)[None, :]*e2[:, None])
    if e1e2:
        ve1 = (-np.sin(psi)[None, :]*nout[:, None] + np.cos(psi)[None, :]*e1[:, None])
        ve2 = np.array([vout[1, :]*ve1[2, :] - vout[2, :]*ve1[1, :],
                        vout[2, :]*ve1[0, :] - vout[0, :]*ve1[2, :],
                        vout[0, :]*ve1[1, :] - vout[1, :]*ve1[0, :]])
        return vout, ve1, ve2
    else:
        return vout

def CrystBragg_sample_outline_plot_sphrect(center, nout, e1, e2,
                                           rcurve, extenthalf, res=None):
    dpsi, dtheta = extenthalf
    if res is None:
        res = np.min(extenthalf)/5.
    npsi = 2*int(np.ceil(dpsi / res)) + 1
    ntheta = 2*int(np.ceil(dtheta / res)) + 1
    psi, theta = CrystBragg_sample_outline_sphrect(dpsi, dtheta,
                                                   npsi=npsi, ntheta=ntheta)
    vout = CrystBragg_get_noute1e2_from_psitheta(nout, e1, e2, psi, theta,
                                                 e1e2=False)
    return center[:, None] + rcurve*vout

def CrystBragg_sample_outline_Rays(center, nout, e1, e2,
                                   rcurve, extenthalf,
                                   bragg, phi):
    dpsi, dtheta = extenthalf
    psi, theta = CrystBragg_sample_outline_sphrect(dpsi, dtheta,
                                                   npsi=3, ntheta=3)
    psi = np.append(psi, [0])
    theta = np.append(theta, [np.pi/2.])
    npts = psi.size

    # add repetitions for rays
    nrays = phi.size
    psi = np.repeat(psi, nrays)
    theta = np.repeat(theta, nrays)

    # add tiling for pts
    bragg = np.tile(bragg, npts)
    phi = np.tile(phi, npts)

    # Compute local vectors
    vout, ve1, ve2 = CrystBragg_get_noute1e2_from_psitheta(nout, e1, e2,
                                                           psi, theta)
    # Deduce D, u
    D = center[:, None] + rcurve*vout
    u = (-np.sin(bragg)*vect
         + np.cos(bragg)*(np.cos(phi)*ve1 + np.sin(phi)*ve2))
    return D, u


# ###############################################
#           lamb <=> bragg
# ###############################################

def get_bragg_from_lamb(lamb, d, n=None):
    """ n*lamb = 2d*sin(bragg) """
    if n is None:
        n = 1
    bragg= np.full(lamb.shape, np.nan)
    sin = n*lamb/(2.*d)
    indok = np.abs(sin) <= 1.
    bragg[indok] = np.arcsin(sin[indok])
    return bragg

def get_lamb_from_bragg(bragg, d, n=None):
    """ n*lamb = 2d*sin(bragg) """
    if n is None:
        n = 1
    return 2*d*np.sin(bragg) / n


# ###############################################
#           Approximate solution
# ###############################################

def get_approx_detector_rel(rcurve, bragg, tangent_to_rowland=None):

    if tangent_to_rowland is None:
        tangent_to_rowland = True

    # distance crystal - det_center
    det_dist = rcurve*np.sin(bragg)

    # det_nout and det_e1 in (nout, e1, e2) (det_e2 = e2)
    n_crystdet_rel = np.r_[-np.sin(bragg), np.cos(bragg), 0.]
    if tangent_to_rowland:
        bragg2 = 2.*bragg
        det_nout_rel = np.r_[-np.cos(bragg2), -np.sin(bragg2), 0.]
        det_ei_rel = np.r_[np.sin(bragg2), -np.cos(bragg2), 0.]
    else:
        det_nout_rel = -n_crystdet_rel
        det_ei_rel = np.r_[np.cos(bragg), np.sin(bragg), 0]
    return det_dist, n_crystdet_rel, det_nout_rel, det_ei_rel


def get_det_abs_from_rel(det_dist, n_crystdet_rel, det_nout_rel, det_ei_rel,
                         summit, nout, e1, e2,
                         ddist=None, di=None, dj=None,
                         dtheta=None, dpsi=None, tilt=None):
    # Reference
    det_nout = (det_nout_rel[0]*nout
                + det_nout_rel[1]*e1 + det_nout_rel[2]*e2)
    det_ei = (det_ei_rel[0]*nout
                + det_ei_rel[1]*e1 + det_ei_rel[2]*e2)
    det_ej = np.cross(det_nout, det_ei)

    # Apply translation of center (ddist, di, dj)
    if ddist is None:
        ddist = 0.
    if di is None:
        di = 0.
    if dj is None:
        dj = 0.
    det_dist += ddist

    n_crystdet = (n_crystdet_rel[0]*nout
                  + n_crystdet_rel[1]*e1 + n_crystdet_rel[2]*e2)
    det_cent = summit + det_dist*n_crystdet + di*det_ei + dj*det_ej

    # Apply angles on unit vectors with respect to themselves
    if dtheta is None:
        dtheta = 0.
    if dpsi is None:
        dpsi = 0.
    if tilt is None:
        tilt = 0.

    # dtheta and dpsi
    det_nout2 = ((np.cos(dpsi)*det_nout
                 + np.sin(dpsi)*det_ei)*np.cos(dtheta)
                 + np.sin(dtheta)*det_ej)
    det_ei2 = (np.cos(dpsi)*det_ei - np.sin(dpsi)*det_nout)
    det_ej2 = np.cross(det_nout2, det_ei2)

    # tilt
    det_ei3 = np.cos(tilt)*det_ei2 + np.sin(tilt)*det_ej2
    det_ej3 = np.cross(det_nout2, det_ei3)

    return det_cent, det_nout2, det_ei3, det_ej3


# ###############################################
#           Coordinates transforms
# ###############################################

def checkformat_vectang(Z, nn, frame_cent, frame_ang):
    # Check / format inputs
    nn = np.atleast_1d(nn).ravel()
    assert nn.size == 3
    nn = nn / np.linalg.norm(nn)
    Z = float(Z)

    frame_cent = np.atleast_1d(frame_cent).ravel()
    assert frame_cent.size == 2
    frame_ang = float(frame_ang)

    return Z, nn, frame_cent, frame_ang


def get_e1e2_detectorplane(nn, nIn):
    e1 = np.cross(nn, nIn)
    e1n = np.linalg.norm(e1)
    if e1n < 1.e-10:
        e1 = np.array([nIn[2], -nIn[1], 0.])
        e1n = np.linalg.norm(e1)
    e1 = e1 / e1n
    e2 = np.cross(nn, e1)
    e2 = e2 / np.linalg.norm(e2)
    return e1, e2


def calc_xixj_from_braggphi(summit, det_cent, det_nout, det_ei, det_ej,
                            nout, e1, e2, bragg, phi):
    sp = (det_cent - summit)
    vect = (-np.sin(bragg)[None, :]*nout[:, None]
            + np.cos(bragg)[None, :]*(np.cos(phi)[None, :]*e1[:, None]
                                      + np.sin(phi)[None, :]*e2[:, None]))
    k = np.sum(sp*det_nout) / np.sum(vect*det_nout[:, None], axis=0)
    pts = summit[:, None] + k[None, :]*vect

    xi = np.sum((pts - det_cent[:, None])*det_ei[:, None], axis=0)
    xj = np.sum((pts - det_cent[:, None])*det_ej[:, None], axis=0)
    return xi, xj


def calc_braggphi_from_xixjpts(det_cent, det_ei, det_ej,
                               summit, nin, e1, e2,
                               xi=None, xj=None, pts=None):

    if pts is None:
        xi = xi[None, ...]
        xj = xj[None, ...]
        if xi.ndim == 1:
            summit = summit[:, None]
            det_cent = det_cent[:, None]
            det_ei, det_ej = det_ei[:, None], det_ej[:, None]
            nin, e1, e2 = nin[:, None], e1[:, None], e2[:, None]
        else:
            summit = summit[:, None, None]
            det_cent = det_cent[:, None, None]
            det_ei, det_ej = det_ei[:, None, None], det_ej[:, None, None]
            nin, e1, e2 = (nin[:, None, None],
                           e1[:, None, None], e2[:, None, None])
        pts = det_cent + xi*det_ei + xj*det_ej
    else:
        assert pts.ndim == 2
        pts = pts[:, :, None]
        summit = summit[:, None, None]
        nin, e1, e2 = nin[:, None, None], e1[:, None, None], e2[:, None, None]

    vect = pts - summit
    vect = vect / np.sqrt(np.sum(vect**2, axis=0))[None, ...]
    bragg = np.arcsin(np.sum(vect*nin, axis=0))

    phi = np.arctan2(np.sum(vect*e2, axis=0), np.sum(vect*e1, axis=0))
    return bragg, phi


def get_lambphifit(lamb, phi, nxi, nxj):
    lambD = lamb.max()-lamb.min()
    lambfit = lamb.min() +lambD*np.linspace(0, 1, nxi)
    phiD = phi.max() - phi.min()
    phifit = phi.min() + phiD*np.linspace(0, 1, nxj)
    return lambfit, phifit


# ###############################################
#           From plasma pts
# ###############################################

def calc_psidthetaphi_from_pts_lamb(pts, center, rcurve,
                                    bragg, nlamb, npts,
                                    nout, e1, e2, extenthalf, ntheta=None):

    if ntheta is None:
        ntheta = 100

    scaPCem = np.full((nlamb, npts), np.nan)
    dtheta = np.full((nlamb, npts, ntheta), np.nan)
    psi = np.full((nlamb, npts, ntheta), np.nan)

    # Get to scalar product
    PC = center[:, None] - pts
    PCnorm2 = np.sum(PC**2, axis=0)
    cos2 = np.cos(bragg)**2
    deltaon4 = (rcurve**2*cos2[:, None]**2
                - (rcurve**2*cos2[:, None]
                   - PCnorm2[None, :]*np.sin(bragg)[:, None]**2))

    # Get two relevant solutions
    ind = deltaon4 >= 0.
    cos2 = np.repeat(cos2[:, None], npts, axis=1)[ind]
    PCnorm = np.tile(np.sqrt(PCnorm2), (nlamb, 1))[ind]
    sol1 = -rcurve*cos2 - np.sqrt(deltaon4[ind])
    sol2 = -rcurve*cos2 + np.sqrt(deltaon4[ind])
    # em is a unit vector and ...
    ind1 = (np.abs(sol1) <= PCnorm) & (sol1 >= -rcurve)
    ind2 = (np.abs(sol2) <= PCnorm) & (sol2 >= -rcurve)
    assert not np.any(ind1 & ind2)
    sol1 = sol1[ind1]
    sol2 = sol2[ind2]
    indn = ind.nonzero()
    ind1 = [indn[0][ind1], indn[1][ind1]]
    ind2 = [indn[0][ind2], indn[1][ind2]]
    scaPCem[ind1[0], ind1[1]] = sol1
    scaPCem[ind2[0], ind2[1]] = sol2
    ind = ~np.isnan(scaPCem)

    # Get equation on PCem
    X = np.sum(PC*nout[:, None], axis=0)
    Y = np.sum(PC*e1[:, None], axis=0)
    Z = np.sum(PC*e2[:, None], axis=0)

    scaPCem = np.repeat(scaPCem[..., None], ntheta, axis=-1)
    ind = ~np.isnan(scaPCem)
    XYnorm = np.repeat(np.repeat(np.sqrt(X**2 + Y**2)[None, :],
                                 nlamb, axis=0)[..., None],
                       ntheta, axis=-1)[ind]
    Z = np.repeat(np.repeat(Z[None, :], nlamb, axis=0)[..., None],
                  ntheta, axis=-1)[ind]
    angextra = np.repeat(
        np.repeat(np.arctan2(Y, X)[None, :], nlamb, axis=0)[..., None],
        ntheta, axis=-1)[ind]
    dtheta[ind] = np.repeat(
        np.repeat(extenthalf[1]*np.linspace(-1, 1, ntheta)[None, :],
                  npts, axis=0)[None, ...],
        nlamb, axis=0)[ind]

    psi[ind] = (np.arccos(
        (scaPCem[ind] - Z*np.sin(dtheta[ind]))/(XYnorm*np.cos(dtheta[ind])))
                + angextra)
    psi[ind] = np.arctan2(np.sin(psi[ind]), np.cos(psi[ind]))
    indnan = (~ind) | (np.abs(psi) > extenthalf[0])
    psi[indnan] = np.nan
    dtheta[indnan] = np.nan
    return dtheta, psi, indnan
